
import os.path
from argparse import Namespace, ArgumentDefaultsHelpFormatter
from contextlib import closing
from datetime import datetime, timezone
from typing import Any, Tuple, Mapping, Dict

from pkg_resources import resource_listdir, resource_stream
from pysasl import AuthenticationCredentials

from pymap.config import IMAPConfig
from pymap.exceptions import InvalidAuth
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.session import LoginProtocol
from pymap.parsing.specials.flag import Flag, Recent

from .filter import FilterSet
from .mailbox import Message, MailboxData, MailboxSet
from ..session import BaseSession

__all__ = ['DictBackend', 'Config', 'Session']


class DictBackend(BackendInterface):
    """Defines a backend that uses an in-memory dictionary for example usage
    and integration testing.

    """

    def __init__(self, login: LoginProtocol, config: IMAPConfig) -> None:
        super().__init__()
        self._login = login
        self._config = config

    @property
    def login(self) -> LoginProtocol:
        return self._login

    @property
    def config(self) -> IMAPConfig:
        return self._config

    @classmethod
    def add_subparser(cls, subparsers) -> None:
        parser = subparsers.add_parser(
            'dict', help='in-memory backend',
            formatter_class=ArgumentDefaultsHelpFormatter)
        parser.add_argument('--demo-data', action='store_true',
                            help='load initial demo data')
        parser.add_argument('--demo-user', default='demouser',
                            metavar='VAL', help='demo user ID')
        parser.add_argument('--demo-password', default='demopass',
                            metavar='VAL', help='demo user password')

    @classmethod
    async def init(cls, args: Namespace) -> Tuple['DictBackend', 'Config']:
        config = Config.from_args(args)
        return cls(Session.login, config), config


class Config(IMAPConfig):
    """The config implementation for the dict backend."""

    def __init__(self, args: Namespace, *, demo_data: bool,
                 demo_user: str, demo_password: str, **extra: Any) -> None:
        super().__init__(args, **extra)
        self._demo_data = demo_data
        self._demo_user = demo_user
        self._demo_password = demo_password
        self.set_cache: Dict[str, Tuple[MailboxSet, FilterSet]] = {}

    @property
    def demo_data(self) -> bool:
        """True if demo data should be loaded at startup."""
        return self._demo_data

    @property
    def demo_user(self) -> str:
        """A login name that is valid at startup, which defaults to
        ``demouser``.

        """
        return self._demo_user

    @property
    def demo_password(self) -> str:
        """The password for the :attr:`.demo_user` login name, which defaults
        to ``demopass``.

        """
        return self._demo_password

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {'demo_data': args.demo_data,
                'demo_user': args.demo_user,
                'demo_password': args.demo_password}


class Session(BaseSession[Message]):
    """The session implementation for the dict backend."""

    resource = __name__

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> Config:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set

    @classmethod
    async def login(cls, credentials: AuthenticationCredentials,
                    config: Config) -> 'Session':
        """Checks the given credentials for a valid login and returns a new
        session. The mailbox data is shared between concurrent and future
        sessions, but only for the lifetime of the process.

        """
        user = credentials.authcid
        password = cls._get_password(config, user)
        if user != credentials.identity:
            raise InvalidAuth()
        elif not credentials.check_secret(password):
            raise InvalidAuth()
        mailbox_set, filter_set = config.set_cache.get(user, (None, None))
        if not mailbox_set or not filter_set:
            mailbox_set = MailboxSet()
            filter_set = FilterSet()
            if config.demo_data:
                await cls._load_demo(mailbox_set, filter_set)
            config.set_cache[user] = (mailbox_set, filter_set)
        return cls(credentials.identity, config, mailbox_set, filter_set)

    @classmethod
    def _get_password(cls, config: Config, user: str) -> str:
        expected_user: str = config.demo_user
        expected_password: str = config.demo_password
        if user == expected_user:
            return expected_password
        raise InvalidAuth()

    @classmethod
    async def _load_demo(cls, mailbox_set: MailboxSet,
                         filter_set: FilterSet) -> None:
        inbox = await mailbox_set.get_mailbox('INBOX')
        await cls._load_demo_mailbox('INBOX', inbox)
        mbx_names = sorted(resource_listdir(cls.resource, 'demo'))
        for name in mbx_names:
            if name == 'sieve':
                await cls._load_demo_sieve(name, filter_set)
            elif name != 'INBOX':
                mbx = await mailbox_set.add_mailbox(name)
                await cls._load_demo_mailbox(name, mbx)

    @classmethod
    async def _load_demo_sieve(cls, name: str, filter_set: FilterSet) -> None:
        path = os.path.join('demo', name)
        sieve_stream = resource_stream(cls.resource, path)
        with closing(sieve_stream):
            sieve = sieve_stream.read()
        await filter_set.put('demo', sieve)
        await filter_set.set_active('demo')

    @classmethod
    async def _load_demo_mailbox(cls, name: str, mbx: MailboxData) -> None:
        path = os.path.join('demo', name)
        msg_names = sorted(resource_listdir(cls.resource, path))
        for msg_name in msg_names:
            if msg_name == '.readonly':
                mbx._readonly = True
                continue
            elif msg_name.startswith('.'):
                continue
            msg_path = os.path.join(path, msg_name)
            message_stream = resource_stream(cls.resource, msg_path)
            with closing(message_stream):
                flags_line = message_stream.readline()
                msg_timestamp = float(message_stream.readline())
                msg_data = message_stream.read()
                msg_dt = datetime.fromtimestamp(msg_timestamp, timezone.utc)
                msg_flags = {Flag(flag) for flag in flags_line.split()}
                if Recent in msg_flags:
                    msg_flags.remove(Recent)
                    msg_recent = True
                else:
                    msg_recent = False
                msg = Message.parse(0, msg_data, msg_flags, msg_dt)
            await mbx.add(msg.append_msg, msg_recent)




import asyncio
from datetime import datetime
from typing import Optional, Sequence, List, Dict, Tuple, FrozenSet, \
    Iterable, Awaitable

from aioredis import Redis, MultiExecError, WatchVariableError  # type: ignore

from pymap.exceptions import MailboxNotFound, MailboxConflict
from pymap.flags import FlagOp
from pymap.interfaces.message import AppendMessage, CachedMessage
from pymap.mailbox import MailboxSnapshot
from pymap.mime import MessageContent
from pymap.parsing.modutf7 import modutf7_encode, modutf7_decode
from pymap.parsing.specials import FetchRequirement, SequenceSet
from pymap.parsing.specials.flag import Flag, Deleted, Seen
from pymap.selected import SelectedSet, SelectedMailbox

from ..mailbox import Message, MailboxDataInterface, MailboxSetInterface

__all__ = ['Message', 'MailboxData', 'MailboxSet']


async def _delete_keys(redis: Redis, prefixes: Iterable[bytes]) -> None:
    for prefix in prefixes:
        cur = b'0'
        match = prefix + b':*'
        while cur:
            cur, keys = await redis.scan(cur, match=match)
            if keys:
                await redis.unlink(*keys)


async def _check_errors(multi) -> bool:
    # Prevents warning about exception never being retrieved.
    errors = await asyncio.gather(*multi._results, return_exceptions=True)
    return any(not isinstance(exc, WatchVariableError) for exc in errors)


class MailboxData(MailboxDataInterface[Message]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxDataInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, name: str, prefix: bytes,
                 uid_validity: int) -> None:
        super().__init__()
        self._redis = redis
        self._prefix = prefix
        self._uid_validity = uid_validity
        self._name = name
        self._selected_set = SelectedSet()

    @property
    def name(self) -> str:
        return self._name

    @property
    def readonly(self) -> bool:
        return False

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    def parse_message(self, append_msg: AppendMessage) -> Message:
        return Message.parse(0, append_msg.message, append_msg.flag_set,
                             append_msg.when, recent=True)

    async def update_selected(self, selected: SelectedMailbox) \
            -> SelectedMailbox:
        selected.uid_validity = self.uid_validity
        last_mod_seq = selected.mod_sequence
        if last_mod_seq is None:
            mod_seq, updated, expunged = await self._get_initial()
        else:
            mod_seq, updated, expunged = await self._get_updated(last_mod_seq)
        selected.mod_sequence = mod_seq
        selected.add_updates(updated, expunged)
        return selected

    async def add(self, append_msg: AppendMessage, recent: bool = False) \
            -> Message:
        redis = self._redis
        prefix = self._prefix
        is_deleted = Deleted in append_msg.flag_set
        is_unseen = Seen not in append_msg.flag_set
        msg_content = MessageContent.parse(append_msg.message)
        msg_flags = [flag.value for flag in append_msg.flag_set]
        msg_time = append_msg.when.isoformat().encode('ascii')
        while True:
            await redis.watch(prefix + b':max-mod')
            max_uid, max_mod = await redis.mget(prefix + b':max-uid',
                                                prefix + b':max-mod')
            new_uid = int(max_uid or 0) + 1
            new_mod = int(max_mod or 0) + 1
            msg_prefix = prefix + b':msg:%d' % new_uid
            multi = redis.multi_exec()
            multi.set(prefix + b':max-uid', new_uid)
            multi.set(prefix + b':max-mod', new_mod)
            multi.sadd(prefix + b':uids', new_uid)
            multi.zadd(prefix + b':mod-sequence', new_mod, new_uid)
            multi.zadd(prefix + b':sequence', new_uid, new_uid)
            if recent:
                multi.sadd(prefix + b':recent', new_uid)
            if is_deleted:
                multi.sadd(prefix + b':deleted', new_uid)
            if is_unseen:
                multi.zadd(prefix + b':unseen', new_uid, new_uid)
            if msg_flags:
                multi.sadd(msg_prefix + b':flags', *msg_flags)
            multi.set(msg_prefix + b':time', msg_time)
            multi.set(msg_prefix + b':header', bytes(msg_content.header))
            multi.set(msg_prefix + b':body', bytes(msg_content.body))
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        return Message(new_uid, append_msg.flag_set, append_msg.when,
                       recent=recent, content=msg_content)

    async def get(self, uid: int, cached_msg: CachedMessage = None,
                  requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> Optional[Message]:
        redis = self._redis
        prefix = self._prefix
        msg_prefix = prefix + b':msg:%d' % uid
        multi = redis.multi_exec()
        multi.sismember(prefix + b':uids', uid)
        multi.smembers(msg_prefix + b':flags')
        multi.get(msg_prefix + b':time')
        multi.sismember(prefix + b':recent', uid)
        if requirement & FetchRequirement.BODY:
            multi.get(msg_prefix + b':header')
            multi.get(msg_prefix + b':body')
        elif requirement & FetchRequirement.HEADERS:
            multi.get(msg_prefix + b':header')
            multi.echo(b'')
        else:
            multi.echo(b'')
            multi.echo(b'')
        exists, flags, time, recent, header, body = await multi.execute()
        if not exists:
            if cached_msg is None:
                return None
            else:
                return Message(cached_msg.uid, cached_msg.permanent_flags,
                               cached_msg.internal_date, expunged=True)
        msg_flags = {Flag(flag) for flag in flags}
        msg_time = datetime.fromisoformat(time.decode('ascii'))
        msg_recent = bool(recent)
        if header:
            msg_content = MessageContent.parse_split(header, body)
            return Message(uid, msg_flags, msg_time, recent=msg_recent,
                           content=msg_content)
        else:
            return Message(uid, msg_flags, msg_time, recent=msg_recent)

    async def delete(self, uids: Iterable[int]) -> None:
        redis = self._redis
        prefix = self._prefix
        uids = list(uids)
        if not uids:
            return
        while True:
            await redis.watch(prefix + b':max-mod')
            max_mod = await redis.get(prefix + b':max-mod')
            new_mod = int(max_mod or 0) + 1
            multi = redis.multi_exec()
            multi.set(prefix + b':max-mod', new_mod)
            for uid in uids:
                multi.zadd(prefix + b':expunged', new_mod, uid)
            multi.srem(prefix + b':uids', *uids)
            multi.zrem(prefix + b':sequence', *uids)
            multi.zrem(prefix + b':mod-sequence', *uids)
            multi.srem(prefix + b':recent', *uids)
            multi.srem(prefix + b':deleted', *uids)
            multi.zrem(prefix + b':unseen', *uids)
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        prefixes = (prefix + b':msg:%d' % uid for uid in uids)
        asyncio.create_task(_delete_keys(redis, prefixes))

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':max-mod')
            recent = await redis.smembers(prefix + b':recent')
            if not recent:
                await redis.unwatch()
                break
            max_mod = await redis.get(prefix + b':max-mod')
            new_mod = int(max_mod or 0) + 1
            multi = self._redis.multi_exec()
            multi.set(prefix + b':max-mod', new_mod)
            for uid in recent:
                multi.zadd(prefix + b':mod-sequence', new_mod, uid)
            multi.delete(prefix + b':recent')
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        for uid_bytes in recent:
            selected.session_flags.add_recent(int(uid_bytes))

    async def update_flags(self, messages: Sequence[Message],
                           flag_set: FrozenSet[Flag], mode: FlagOp) -> None:
        redis = self._redis
        prefix = self._prefix
        messages = list(messages)
        if not messages:
            return
        uids = {msg.uid: msg for msg in messages}
        while True:
            await redis.watch(prefix + b':max-mod')
            pipe = redis.pipeline()
            pipe.smembers(prefix + b':uids')
            pipe.get(prefix + b':max-mod')
            existing_uids, max_mod = await pipe.execute()
            update_uids = uids.keys() & {int(uid) for uid in existing_uids}
            if not update_uids:
                await redis.unwatch()
                break
            new_mod = int(max_mod or 0) + 1
            new_flags: Dict[int, Awaitable[Sequence[bytes]]] = {}
            multi = redis.multi_exec()
            multi.set(prefix + b':max-mod', new_mod)
            for msg in messages:
                msg_uid = msg.uid
                if msg_uid not in update_uids:
                    continue
                msg_prefix = prefix + b':msg:%d' % msg_uid
                flag_vals = (flag.value for flag in flag_set)
                multi.zadd(prefix + b':mod-sequence', new_mod, msg_uid)
                if mode == FlagOp.REPLACE:
                    multi.unlink(msg_prefix + b':flags')
                    if flag_set:
                        multi.sadd(msg_prefix + b':flags', *flag_vals)
                    if Deleted in flag_set:
                        multi.sadd(prefix + b':deleted', msg_uid)
                    else:
                        multi.srem(prefix + b':deleted', msg_uid)
                    if Seen not in flag_set:
                        multi.zadd(prefix + b':unseen', msg_uid, msg_uid)
                    else:
                        multi.zrem(prefix + b':unseen', msg_uid)
                elif mode == FlagOp.ADD and flag_set:
                    multi.sadd(msg_prefix + b':flags', *flag_vals)
                    if Deleted in flag_set:
                        multi.sadd(prefix + b':deleted', msg_uid)
                    if Seen in flag_set:
                        multi.zrem(prefix + b':unseen', msg_uid)
                elif mode == FlagOp.DELETE and flag_set:
                    multi.srem(msg_prefix + b':flags', *flag_vals)
                    if Deleted in flag_set:
                        multi.srem(prefix + b':deleted', msg_uid)
                    if Seen in flag_set:
                        multi.zadd(prefix + b':unseen', msg_uid, msg_uid)
                new_flags[msg_uid] = multi.smembers(msg_prefix + b':flags')
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                for msg_uid, msg_flags in new_flags.items():
                    msg = uids[msg_uid]
                    msg.permanent_flags = frozenset(
                        Flag(flag) for flag in await msg_flags)
                break

    async def cleanup(self) -> None:
        pass

    async def find_deleted(self, seq_set: SequenceSet,
                           selected: SelectedMailbox) -> Sequence[int]:
        if not seq_set.is_all:
            return await super().find_deleted(seq_set, selected)
        redis = self._redis
        prefix = self._prefix
        deleted = await redis.smembers(prefix + b':deleted')
        return [int(uid) for uid in deleted]

    async def snapshot(self) -> MailboxSnapshot:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':sequence')
            pipe = redis.pipeline()
            pipe.get(prefix + b':max-uid')
            pipe.zcard(prefix + b':sequence')
            pipe.scard(prefix + b':recent')
            pipe.zcard(prefix + b':unseen')
            pipe.zrange(prefix + b':unseen', 0, 0)
            max_uid, exists, num_recent, num_unseen, unseen = \
                await pipe.execute()
            next_uid = int(max_uid or 0) + 1
            if not unseen:
                await redis.unwatch()
                first_unseen: Optional[int] = None
                break
            else:
                first_uid = int(unseen[0])
                multi = redis.multi_exec()
                multi.zrank(prefix + b':sequence', first_uid)
                try:
                    [first_unseen] = await multi.execute()
                except MultiExecError:
                    if await _check_errors(multi):
                        raise
                else:
                    break
        return MailboxSnapshot(self.name, self.readonly, self.uid_validity,
                               self.permanent_flags, self.session_flags,
                               exists, num_recent, num_unseen, first_unseen,
                               next_uid)

    async def _get_initial(self) \
            -> Tuple[int, Sequence[Message], Sequence[int]]:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':max-mod')
            uids = await redis.zrange(prefix + b':sequence')
            multi = redis.multi_exec()
            multi.get(prefix + b':max-mod')
            for uid in uids:
                msg_prefix = prefix + b':msg:' + uid
                multi.echo(uid)
                multi.smembers(msg_prefix + b':flags')
                multi.get(msg_prefix + b':time')
            try:
                results = await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        mod_seq = int(results[0] or 0)
        updated: List[Message] = []
        for i in range(1, len(results), 3):
            msg_uid = int(results[i])
            msg_flags = {Flag(flag) for flag in results[i + 1]}
            msg_time = datetime.fromisoformat(results[i + 2].decode('ascii'))
            msg = Message(msg_uid, msg_flags, msg_time)
            updated.append(msg)
        return mod_seq, updated, []

    async def _get_updated(self, last_mod_seq: int) \
            -> Tuple[int, Sequence[Message], Sequence[int]]:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':max-mod')
            uids = await redis.zrangebyscore(
                prefix + b':mod-sequence', last_mod_seq)
            multi = redis.multi_exec()
            multi.get(prefix + b':max-mod')
            multi.zrangebyscore(prefix + b':expunged', last_mod_seq)
            for uid in uids:
                msg_prefix = prefix + b':msg:' + uid
                multi.echo(uid)
                multi.smembers(msg_prefix + b':flags')
                multi.get(msg_prefix + b':time')
            try:
                results = await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        mod_seq = int(results[0] or 0)
        expunged = [int(uid) for uid in results[1]]
        updated: List[Message] = []
        for i in range(2, len(results), 3):
            msg_uid = int(results[i])
            msg_flags = {Flag(flag) for flag in results[i + 1]}
            msg_time = datetime.fromisoformat(results[i + 2].decode('ascii'))
            msg = Message(msg_uid, msg_flags, msg_time)
            updated.append(msg)
        return mod_seq, updated, expunged


class MailboxSet(MailboxSetInterface[MailboxData]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxSetInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, prefix: bytes) -> None:
        super().__init__()
        self._redis = redis
        self._prefix = prefix
        self._order_key = prefix + b':mbx-order'
        self._mbx_key = prefix + b':mailboxes'
        self._sub_key = prefix + b':subscribed'
        self._uidv_key = prefix + b':uid-validity'

    @property
    def delimiter(self) -> str:
        return '/'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        if subscribed:
            await self._redis.sadd(self._sub_key, name)
        else:
            await self._redis.srem(self._sub_key, name)

    async def list_subscribed(self) -> Sequence[str]:
        return [modutf7_decode(name) for name in
                await self._redis.smembers(self._sub_key)]

    async def list_mailboxes(self) -> Sequence[str]:
        return [modutf7_decode(name) for name in
                await self._redis.zrange(self._mbx_key)]

    async def get_mailbox(self, name: str,
                          try_create: bool = False) -> 'MailboxData':
        redis = self._redis
        name_key = modutf7_encode(name)
        multi = redis.multi_exec()
        multi.zscore(self._mbx_key, name_key)
        multi.hget(self._uidv_key, name_key)
        exists, uidval = await multi.execute()
        if not exists:
            raise MailboxNotFound(name, try_create)
        mbx_prefix = b':'.join((self._prefix, name_key, uidval))
        return MailboxData(redis, name, mbx_prefix, int(uidval))

    async def add_mailbox(self, name: str) -> 'MailboxData':
        redis = self._redis
        name_key = modutf7_encode(name)
        while True:
            await redis.watch(self._mbx_key)
            exists = await redis.zscore(self._mbx_key, name_key)
            if exists:
                await redis.unwatch()
                raise MailboxConflict(name)
            order = await redis.incr(self._order_key)
            multi = redis.multi_exec()
            multi.hincrby(self._uidv_key, name_key)
            multi.zadd(self._mbx_key, order, name_key)
            try:
                uidval, _ = await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                mbx_prefix = b':'.join(
                    (self._prefix, name_key, b'%d' % uidval))
                return MailboxData(redis, name, mbx_prefix, uidval)

    async def delete_mailbox(self, name: str) -> None:
        redis = self._redis
        name_key = modutf7_encode(name)
        multi = redis.multi_exec()
        multi.hget(self._uidv_key, name_key)
        multi.zrem(self._mbx_key, name_key)
        uidval, exists = await multi.execute()
        if not exists:
            raise MailboxNotFound(name)
        mbx_prefix = b':'.join((self._prefix, name_key, uidval))
        asyncio.create_task(_delete_keys(redis, [mbx_prefix]))

    async def rename_mailbox(self, before: str, after: str) -> 'MailboxData':
        raise NotImplementedError()  # TODO



"""Implements some concurrency utilities used by pymap. Each has
an implementation using :mod:`asyncio` and :mod:`threading` concurrency
primitives.

"""

import asyncio
import os.path
import time
from abc import abstractmethod, ABCMeta
from asyncio import Event as _asyncio_Event, Lock as _asyncio_Lock
from concurrent.futures import Executor, ThreadPoolExecutor, TimeoutError
from contextlib import asynccontextmanager
from contextvars import copy_context, Context
from threading import local, Event as _threading_Event, Lock as _threading_Lock
from typing import cast, TypeVar, Optional, Sequence, MutableSet, Awaitable, \
    AsyncContextManager, AsyncIterator
from weakref import WeakSet

__all__ = ['Subsystem', 'Event', 'ReadWriteLock', 'FileLock', 'EventT', 'RetT']

#: Type variable with an upper bound of :class:`Event`.
EventT = TypeVar('EventT', bound='Event')

#: Type variable for any return type.
RetT = TypeVar('RetT')

_Delay = Sequence[float]


class Subsystem(metaclass=ABCMeta):
    """Utility for creating concurrency primitives for a subsystem, either
    :mod:`asyncio` or :mod:`threading`.

    """

    @classmethod
    def for_executor(cls, executor: Optional[Executor]) -> 'Subsystem':
        """Return a subsystem based on the given executor. If ``executor`` is
        None, use :mod:`asyncio`. If ``executor`` is a
        :class:`concurrent.futures.ThreadPoolExecutor`, use :mod:`threading`.

        Args:
            executor: The executor in use, if any.

        """
        if isinstance(executor, ThreadPoolExecutor):
            return _ThreadingSubsystem(executor)
        elif executor is None:
            return _AsyncioSubsystem()
        else:
            raise TypeError(executor)

    @classmethod
    def for_asyncio(cls) -> 'Subsystem':
        """Return a subsystem for :mod:`asyncio`."""
        return _AsyncioSubsystem()

    @classmethod
    def for_threading(cls, executor: ThreadPoolExecutor) -> 'Subsystem':
        """Return a subsystem for :mod:`threading`."""
        return _ThreadingSubsystem(executor)

    @property
    @abstractmethod
    def subsystem(self) -> str:
        """The subsystem name, ``'asyncio'`` or ``'threading'``."""
        ...

    @abstractmethod
    def execute(self, future: Awaitable[RetT]) -> Awaitable[RetT]:
        """Executes the future and returns its result in the subsystem. For
        :mod:`asyncio`, this simply means ``return await future``. For
        :mod:`threading`, it uses
        :meth:`~asyncio.AbstractEventLoop.run_in_executor` to run the future in
        another thread.

        Args:
            future: An awaitable result to execute by the subsystem.

        """
        ...

    @abstractmethod
    def new_rwlock(self) -> 'ReadWriteLock':
        """Return a new read-write lock."""
        ...

    @abstractmethod
    def new_event(self) -> 'Event':
        """Return a new concurrent event."""
        ...


class _AsyncioSubsystem(Subsystem):

    @property
    def subsystem(self) -> str:
        return 'asyncio'

    def execute(self, future: Awaitable[RetT]) -> Awaitable[RetT]:
        return future

    def new_rwlock(self) -> '_AsyncioReadWriteLock':
        return _AsyncioReadWriteLock()

    def new_event(self) -> '_AsyncioEvent':
        return _AsyncioEvent()


class _ThreadingSubsystem(Subsystem):  # pragma: no cover

    class _EventLoopLocal(local):

        def __init__(self) -> None:
            self.event_loop = asyncio.new_event_loop()

    def __init__(self, executor: ThreadPoolExecutor) -> None:
        super().__init__()
        self._local = self._EventLoopLocal()
        self._executor = executor

    @property
    def subsystem(self) -> str:
        return 'threading'

    async def execute(self, future: Awaitable[RetT]) -> RetT:
        loop = asyncio.get_event_loop()
        ctx = copy_context()
        return await loop.run_in_executor(
            self._executor, self._run_in_thread, future, ctx)

    def _run_in_thread(self, future: Awaitable[RetT], ctx: Context) -> RetT:
        loop = self._local.event_loop
        ret = ctx.run(loop.run_until_complete, future)
        return cast(RetT, ret)

    def new_rwlock(self) -> '_ThreadingReadWriteLock':
        return _ThreadingReadWriteLock()

    def new_event(self) -> '_ThreadingEvent':
        return _ThreadingEvent()


class ReadWriteLock(metaclass=ABCMeta):
    """Read-write lock."""

    @classmethod
    def for_asyncio(cls) -> 'ReadWriteLock':
        """Return a read-write lock for asyncio."""
        return _AsyncioReadWriteLock()

    @classmethod
    def for_threading(cls) -> 'ReadWriteLock':
        """Return a read-write lock for threading."""
        return _ThreadingReadWriteLock()

    @property
    @abstractmethod
    def subsystem(self) -> str:
        """The subsystem the read-write lock was created for, ``'asyncio'`` or
        ``'threading'``.

        """
        ...

    @abstractmethod
    def read_lock(self) -> AsyncContextManager[None]:
        """Acquires a read-lock, blocking until any write-locks are released.

        """
        ...

    @abstractmethod
    def write_lock(self) -> AsyncContextManager[None]:
        """Acquires a write-lock, blocking until all read- or write-locks are
        released.

        """
        ...


class Event(metaclass=ABCMeta):
    """Concurrent event, one thread signals and any waiting event is released.

    """

    @classmethod
    def for_asyncio(cls) -> 'Event':
        """Return an event for asyncio."""
        return _AsyncioEvent()

    @classmethod
    def for_threading(cls) -> 'Event':
        """Return an event for threading."""
        return _ThreadingEvent()

    @property
    @abstractmethod
    def subsystem(self) -> str:
        """The subsystem the event was created for, ``'asyncio'`` or
        ``'threading'``.

        """
        ...

    @abstractmethod
    def or_event(self: EventT, *events: EventT) -> EventT:
        """Return a new event that is signalled when either the current event
        or any of the provided events are signalled.

        """
        ...

    @abstractmethod
    def is_set(self) -> bool:
        """Return True if the event is set."""
        ...

    @abstractmethod
    def set(self) -> None:
        """Signal the waiting threads to release."""
        ...

    @abstractmethod
    def clear(self) -> None:
        """Clear the signal, allowing threads to wait again."""
        ...

    @abstractmethod
    async def wait(self) -> None:
        """Wait until another thread signals the event."""
        ...


class _AsyncioReadWriteLock(ReadWriteLock):

    def __init__(self) -> None:
        super().__init__()
        self._read_lock = _asyncio_Lock()
        self._write_lock = _asyncio_Lock()
        self._counter = 0

    @property
    def subsystem(self) -> str:
        return 'asyncio'

    async def _acquire_read(self) -> bool:
        async with self._read_lock:
            self._counter += 1
            return self._counter == 1

    async def _release_read(self) -> bool:
        async with self._read_lock:
            self._counter -= 1
            return self._counter == 0

    @asynccontextmanager
    async def read_lock(self) -> AsyncIterator[None]:
        if await self._acquire_read():
            await self._write_lock.acquire()
        try:
            yield
        finally:
            if await self._release_read():
                self._write_lock.release()

    @asynccontextmanager
    async def write_lock(self) -> AsyncIterator[None]:
        async with self._write_lock:
            yield


class _ThreadingReadWriteLock(ReadWriteLock):  # pragma: no cover

    def __init__(self) -> None:
        super().__init__()
        self._read_lock = _threading_Lock()
        self._write_lock = _threading_Lock()
        self._counter = 0

    @property
    def subsystem(self) -> str:
        return 'threading'

    def _acquire_read(self) -> bool:
        with self._read_lock:
            self._counter += 1
            return self._counter == 1

    def _release_read(self) -> bool:
        with self._read_lock:
            self._counter -= 1
            return self._counter == 0

    @asynccontextmanager
    async def read_lock(self) -> AsyncIterator[None]:
        if self._acquire_read():
            self._write_lock.acquire()
        try:
            yield
        finally:
            if self._release_read():
                self._write_lock.release()

    @asynccontextmanager
    async def write_lock(self) -> AsyncIterator[None]:
        with self._write_lock:
            yield


class FileLock(ReadWriteLock):  # pragma: no cover
    """Uses the presence or absence of a file on the filesystem to simulate
    a read-write lock. If the file is present, other read- and write-locks will
    block until the file is gone. If the file is absent, read-locks will not
    block. Write-locks will create the file on acquire and remove it on
    release.

    The delay arguments are a sequence of floats used as the duration of
    successive :func:`~asyncio.sleep` calls. If this sequence is exhausted
    before a lock is established, :class:`TimeoutError` is thrown.

    Args:
        path: The path of the lock file.
        expiration: Lock files older than this age will be deleted.
        read_retry_delay: The delay sequence between read-lock attempts.
        write_retry_delay: The delay sequence between write-lock attempts.

    """

    _DEFAULT_DELAY = (0.01, 0.03, 0.06, 0.1, 0.15, 0.25, 0.4, 0.5, 0.5, 1.0,
                      1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0)

    def __init__(self, path: str, expiration: float = 600.0,
                 read_retry_delay: _Delay = _DEFAULT_DELAY,
                 write_retry_delay: _Delay = _DEFAULT_DELAY) \
            -> None:
        super().__init__()
        self._path = path
        self._expiration = expiration
        self._read_retry_delay = read_retry_delay
        self._write_retry_delay = write_retry_delay

    @property
    def subsystem(self) -> str:
        """The subsystem the read-write lock was created for, ``'file'``."""
        return 'file'

    def _check_lock(self) -> bool:
        try:
            statinfo = os.stat(self._path)
        except FileNotFoundError:
            return True
        else:
            if time.time() - statinfo.st_mtime >= self._expiration:
                try:
                    os.unlink(self._path)
                except OSError:
                    pass
                return True
            return False

    def _try_lock(self) -> bool:
        try:
            with open(self._path, 'x'):
                pass
        except FileExistsError:
            return False
        else:
            return True

    def _unlock(self) -> None:
        try:
            os.unlink(self._path)
        except OSError:
            pass

    @asynccontextmanager
    async def read_lock(self) -> AsyncIterator[None]:
        if self._check_lock():
            yield
            return
        for delay in self._read_retry_delay:
            await asyncio.sleep(delay)
            if not os.path.exists(self._path):
                yield
                break
        else:
            raise TimeoutError()

    @asynccontextmanager
    async def write_lock(self) -> AsyncIterator[None]:
        if self._check_lock() and self._try_lock():
            try:
                yield
            finally:
                self._unlock()
            return
        for delay in self._write_retry_delay:
            await asyncio.sleep(delay)
            if self._try_lock():
                try:
                    yield
                finally:
                    self._unlock()
                break
        else:
            raise TimeoutError()


class _AsyncioEvent(Event):

    def __init__(self) -> None:
        super().__init__()
        self._event = _asyncio_Event()
        self._listeners: MutableSet['_AsyncioEvent'] = WeakSet()

    @property
    def subsystem(self) -> str:
        return 'asyncio'

    def or_event(self, *events: '_AsyncioEvent') -> '_AsyncioEvent':
        or_event = _AsyncioEvent()
        self._listeners.add(or_event)
        for event in events:
            event._listeners.add(or_event)
        return or_event

    def is_set(self) -> bool:
        return self._event.is_set()

    def set(self) -> None:
        self._event.set()
        for listener in self._listeners:
            listener.set()

    def clear(self) -> None:
        self._event.clear()

    async def wait(self) -> None:
        await self._event.wait()


class _ThreadingEvent(Event):  # pragma: no cover

    def __init__(self) -> None:
        super().__init__()
        self._event = _threading_Event()
        self._listeners: MutableSet['_ThreadingEvent'] = WeakSet()

    @property
    def subsystem(self) -> str:
        return 'threading'

    def or_event(self, *events: '_ThreadingEvent') -> '_ThreadingEvent':
        or_event = _ThreadingEvent()
        self._listeners.add(or_event)
        for event in events:
            event._listeners.add(or_event)
        return or_event

    def is_set(self) -> bool:
        return self._event.is_set()

    def set(self) -> None:
        self._event.set()
        for listener in self._listeners:
            listener.set()

    def clear(self) -> None:
        self._event.clear()

    async def wait(self) -> None:
        self._event.wait()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymap/admin/grpc/admin.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pymap/admin/grpc/admin.proto',
  package='admin',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1cpymap/admin/grpc/admin.proto\x12\x05\x61\x64min\"|\n\rAppendRequest\x12\x0c\n\x04user\x18\x01 \x01(\t\x12\x0e\n\x06sender\x18\x02 \x01(\t\x12\x11\n\trecipient\x18\x03 \x01(\t\x12\x0f\n\x07mailbox\x18\x04 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x12\r\n\x05\x66lags\x18\x06 \x03(\t\x12\x0c\n\x04when\x18\x07 \x01(\x04\"\x8b\x01\n\x0e\x41ppendResponse\x12\x1d\n\x06result\x18\x01 \x01(\x0e\x32\r.admin.Result\x12\x12\n\nerror_type\x18\x02 \x01(\t\x12\x16\n\x0e\x65rror_response\x18\x03 \x01(\x0c\x12\x0f\n\x07mailbox\x18\x04 \x01(\t\x12\x10\n\x08validity\x18\x05 \x01(\r\x12\x0b\n\x03uid\x18\x06 \x01(\r*)\n\x06Result\x12\x0b\n\x07SUCCESS\x10\x00\x12\x12\n\x0e\x45RROR_RESPONSE\x10\x01\x32@\n\x05\x41\x64min\x12\x37\n\x06\x41ppend\x12\x14.admin.AppendRequest\x1a\x15.admin.AppendResponse\"\x00\x62\x06proto3')
)

_RESULT = _descriptor.EnumDescriptor(
  name='Result',
  full_name='admin.Result',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR_RESPONSE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=307,
  serialized_end=348,
)
_sym_db.RegisterEnumDescriptor(_RESULT)

Result = enum_type_wrapper.EnumTypeWrapper(_RESULT)
SUCCESS = 0
ERROR_RESPONSE = 1



_APPENDREQUEST = _descriptor.Descriptor(
  name='AppendRequest',
  full_name='admin.AppendRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='admin.AppendRequest.user', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sender', full_name='admin.AppendRequest.sender', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='recipient', full_name='admin.AppendRequest.recipient', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mailbox', full_name='admin.AppendRequest.mailbox', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='admin.AppendRequest.data', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flags', full_name='admin.AppendRequest.flags', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='when', full_name='admin.AppendRequest.when', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39,
  serialized_end=163,
)


_APPENDRESPONSE = _descriptor.Descriptor(
  name='AppendResponse',
  full_name='admin.AppendResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='admin.AppendResponse.result', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_type', full_name='admin.AppendResponse.error_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_response', full_name='admin.AppendResponse.error_response', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mailbox', full_name='admin.AppendResponse.mailbox', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='validity', full_name='admin.AppendResponse.validity', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uid', full_name='admin.AppendResponse.uid', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=305,
)

_APPENDRESPONSE.fields_by_name['result'].enum_type = _RESULT
DESCRIPTOR.message_types_by_name['AppendRequest'] = _APPENDREQUEST
DESCRIPTOR.message_types_by_name['AppendResponse'] = _APPENDRESPONSE
DESCRIPTOR.enum_types_by_name['Result'] = _RESULT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AppendRequest = _reflection.GeneratedProtocolMessageType('AppendRequest', (_message.Message,), dict(
  DESCRIPTOR = _APPENDREQUEST,
  __module__ = 'pymap.admin.grpc.admin_pb2'
  # @@protoc_insertion_point(class_scope:admin.AppendRequest)
  ))
_sym_db.RegisterMessage(AppendRequest)

AppendResponse = _reflection.GeneratedProtocolMessageType('AppendResponse', (_message.Message,), dict(
  DESCRIPTOR = _APPENDRESPONSE,
  __module__ = 'pymap.admin.grpc.admin_pb2'
  # @@protoc_insertion_point(class_scope:admin.AppendResponse)
  ))
_sym_db.RegisterMessage(AppendResponse)



_ADMIN = _descriptor.ServiceDescriptor(
  name='Admin',
  full_name='admin.Admin',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=350,
  serialized_end=414,
  methods=[
  _descriptor.MethodDescriptor(
    name='Append',
    full_name='admin.Admin.Append',
    index=0,
    containing_service=None,
    input_type=_APPENDREQUEST,
    output_type=_APPENDRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_ADMIN)

DESCRIPTOR.services_by_name['Admin'] = _ADMIN

# @@protoc_insertion_point(module_scope)




import re
from abc import abstractmethod, ABCMeta
from email.headerregistry import Address, AddressHeader
from typing import Sequence, List, Pattern

from pymap.interfaces.message import AppendMessage
from pymap.mime import MessageContent
from sievelib.commands import Command  # type: ignore

from .util import AddressPart, MatchType, SizeComparator, str_list

__all__ = ['SieveTest']


class SieveTest(metaclass=ABCMeta):
    """Base class for all sieve tests.

    See Also:
        `RFC 5228 5. <https://tools.ietf.org/html/rfc5228#section-5>`_

    """

    @classmethod
    def of(cls, cmd: Command) -> 'SieveTest':
        """Given a test command, determine the implementation sub-class and
        return an instance.

        Args:
            cmd: The test command.

        """
        if cmd.name == 'allof':
            return AllOfTest([SieveTest.of(test)
                              for test in cmd.arguments['tests']])
        elif cmd.name == 'anyof':
            return AnyOfTest([SieveTest.of(test)
                              for test in cmd.arguments['tests']])
        elif cmd.name == 'not':
            return NotTest(SieveTest.of(cmd.arguments['test']))
        elif cmd.name == 'true':
            return TrueTest()
        elif cmd.name == 'false':
            return FalseTest()
        elif cmd.name in ('address', 'envelope'):
            address_part = AddressPart.of(cmd.arguments.get('address-part'))
            match_type = MatchType.of(cmd.arguments.get('match-type'))
            case_sensitive = cmd.arguments.get('comparator') == '"i;octet"'
            header_list = str_list(cmd.arguments['header-list'])
            key_list = str_list(cmd.arguments['key-list'])
            if cmd.name == 'address':
                return AddressTest(header_list, key_list, address_part,
                                   match_type, case_sensitive)
            else:
                return EnvelopeTest(header_list, key_list, address_part,
                                    match_type, case_sensitive)
        elif cmd.name == 'exists':
            header_names = str_list(cmd.arguments['header-names'])
            return ExistsTest(header_names)
        elif cmd.name == 'header':
            match_type = MatchType.of(cmd.arguments.get('match-type'))
            case_sensitive = cmd.arguments.get('comparator') == '"i;octet"'
            header_names = str_list(cmd.arguments['header-names'])
            key_list = str_list(cmd.arguments['key-list'])
            return HeaderTest(header_names, key_list, match_type,
                              case_sensitive)
        elif cmd.name == 'size':
            comparator = SizeComparator.of(cmd.arguments['comparator'])
            limit = int(cmd.arguments['limit'])
            return SizeTest(comparator, limit)
        else:
            raise NotImplementedError(cmd.name)

    @abstractmethod
    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        """Run the test implementation on the appended message data, returning
        True if the test matched the data. Must be overridden by sub-classes.

        Args:
            sender: The envelope sender of the message.
            recipient: The envelope recipient of the message.
            append_msg: The message to be appended.
            content: The parsed message content.

        """
        ...


class AllOfTest(SieveTest):

    def __init__(self, tests: Sequence[SieveTest]) -> None:
        super().__init__()
        self._tests = tests

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return all(test.run(sender, recipient, append_msg, content)
                   for test in self._tests)


class AnyOfTest(SieveTest):

    def __init__(self, tests: Sequence[SieveTest]) -> None:
        super().__init__()
        self._tests = tests

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return any(test.run(sender, recipient, append_msg, content)
                   for test in self._tests)


class NotTest(SieveTest):

    def __init__(self, test: SieveTest) -> None:
        super().__init__()
        self._test = test

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return not self._test.run(sender, recipient, append_msg, content)


class TrueTest(SieveTest):

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return True


class FalseTest(SieveTest):

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return False


class MatchTest(SieveTest):

    _wildcards = re.compile(r'([\*\?])')

    @classmethod
    def _compile(cls, key_list: Sequence[str], match_type: MatchType,
                 case_sensitive: bool) -> Sequence[Pattern]:
        flags = 0 if case_sensitive else re.I
        ret: List[Pattern] = []
        for key in key_list:
            if match_type == MatchType.MATCHES:
                pattern_parts: List[str] = []
                for part in cls._wildcards.split(key):
                    if part == '*':
                        pattern_parts.append('.*?')
                    elif part == '?':
                        pattern_parts.append('.')
                    else:
                        pattern_parts.append(re.escape(part))
                pattern = ''.join(pattern_parts)
            else:
                pattern = re.escape(key)
            if match_type in (MatchType.IS, MatchType.MATCHES):
                pattern = '^' + pattern + '$'
            ret.append(re.compile(pattern, flags))
        return ret


class AddressTest(MatchTest):

    def __init__(self, header_list: Sequence[str], key_list: Sequence[str],
                 address_part: AddressPart, match_type: MatchType,
                 case_sensitive: bool) -> None:
        super().__init__()
        self._header_set = {hdr.encode('ascii') for hdr in header_list}
        self._address_part = address_part
        self._patterns = self._compile(key_list, match_type, case_sensitive)

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        part = self._address_part
        for name in self._header_set:
            values = content.header.parsed.get(name, [])
            for value in values:
                if isinstance(value, AddressHeader):
                    for address in value.addresses:
                        if part == AddressPart.LOCALPART:
                            addr_part = address.username
                        elif part == AddressPart.DOMAIN:
                            addr_part = address.domain
                        else:
                            addr_part = address.addr_spec
                        for pattern in self._patterns:
                            if pattern.match(addr_part) is not None:
                                return True
        return False


class EnvelopeTest(AddressTest):

    def _check_address(self, envelope_part: bytes, address: str) -> bool:
        if envelope_part not in self._header_set:
            return False
        try:
            addr_obj = Address(addr_spec=address)
        except Exception:
            addr_obj = Address(username=address)
        part = self._address_part
        if part == AddressPart.LOCALPART:
            addr_part = addr_obj.username
        elif part == AddressPart.DOMAIN:
            addr_part = addr_obj.domain
        else:
            addr_part = addr_obj.addr_spec
            if addr_part == '<>':
                addr_part = ''
        for pattern in self._patterns:
            if pattern.match(addr_part) is not None:
                return True
        return False

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        if self._check_address(b'from', sender):
            return True
        elif self._check_address(b'to', recipient):
            return True
        else:
            return False


class ExistsTest(SieveTest):

    def __init__(self, header_names: Sequence[str]) -> None:
        super().__init__()
        self._header_names = {hdr.encode('ascii') for hdr in header_names}

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return all(name in content.header.parsed
                   for name in self._header_names)


class HeaderTest(MatchTest):

    def __init__(self, header_names: Sequence[str], key_list: Sequence[str],
                 match_type: MatchType, case_sensitive: bool) -> None:
        super().__init__()
        self._header_names = [hdr.encode('ascii') for hdr in header_names]
        self._patterns = self._compile(key_list, match_type, case_sensitive)

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        for name in self._header_names:
            values = content.header.parsed.get(name, [])
            for value in values:
                value_str = str(value)
                for pattern in self._patterns:
                    if pattern.match(value_str) is not None:
                        return True
        return False


class SizeTest(SieveTest):

    def __init__(self, comparator: SizeComparator, limit: int) -> None:
        super().__init__()
        self._comparator = comparator
        self._limit = limit

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        length = len(content)
        if self._comparator == SizeComparator.OVER:
            return length > self._limit
        else:
            return length < self._limit




import unittest

from pymap.parsing import Params
from pymap.parsing.command import Command, CommandNoArgs
from pymap.parsing.commands import Commands


class TestCommandNoArgs(unittest.TestCase):

    def test_parse(self):
        ret, buf = CommandNoArgs.parse(b'    \r\n test', Params())
        self.assertIsInstance(ret, CommandNoArgs)
        self.assertEqual(b' test', buf)

    def test_parse_no_cr(self):
        ret, buf = CommandNoArgs.parse(b'    \n test', Params())
        self.assertIsInstance(ret, CommandNoArgs)
        self.assertEqual(b' test', buf)


class TestCommands(unittest.TestCase):

    def setUp(self):
        self.commands = Commands()

    def test_parse(self):
        cmd, buf = self.commands.parse(b'a0 NOOP\n  ', Params())
        self.assertIsInstance(cmd, Command)
        self.assertEqual(b'a0', cmd.tag)
        self.assertEqual(b'  ', buf)



"""Contains the package version string.

See Also:
    `PEP 396 <https://www.python.org/dev/peps/pep-0396/>`_

"""

import pkg_resources

__all__ = ['__version__']

#: The package version string.
__version__: str = pkg_resources.require(__package__)[0].version




import hashlib
import json
from argparse import Namespace, ArgumentDefaultsHelpFormatter
from typing import Any, Optional, Tuple, Mapping

from aioredis import create_redis, Redis  # type: ignore
from pysasl import AuthenticationCredentials

from pymap.config import IMAPConfig
from pymap.exceptions import InvalidAuth, MailboxConflict
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.session import LoginProtocol

from .filter import FilterSet
from .mailbox import Message, MailboxSet
from ..session import BaseSession

try:
    from passlib.apps import ldap_context  # type: ignore
except ImportError:
    ldap_context = None

__all__ = ['RedisBackend', 'Config', 'Session']


class RedisBackend(BackendInterface):
    """Defines a backend that uses redis data structures for mailbox storage.

    """

    def __init__(self, login: LoginProtocol, config: IMAPConfig) -> None:
        super().__init__()
        self._login = login
        self._config = config

    @property
    def login(self) -> LoginProtocol:
        return self._login

    @property
    def config(self) -> IMAPConfig:
        return self._config

    @classmethod
    def add_subparser(cls, subparsers) -> None:
        parser = subparsers.add_parser(
            'redis', help='redis backend',
            formatter_class=ArgumentDefaultsHelpFormatter)
        parser.add_argument('address', nargs='?', default='redis://localhost',
                            help='the redis server address')
        parser.add_argument('--select', metavar='DB', type=int,
                            help='the redis database for mail data')
        parser.add_argument('--prefix', metavar='VAL', default='',
                            help='the prefix for redis keys')
        parser.add_argument('--users-hash', metavar='KEY',
                            help='the hash key for user lookup')
        parser.add_argument('--users-key', metavar='FORMAT', default='{name}',
                            help='the user lookup key template')
        parser.add_argument('--users-json', action='store_true',
                            help='the user lookup value contains JSON')

    @classmethod
    async def init(cls, args: Namespace) -> Tuple['RedisBackend', 'Config']:
        config = Config.from_args(args)
        return cls(Session.login, config), config


class Config(IMAPConfig):
    """The config implementation for the redis backend.

    Args:
        args: The command-line arguments.
        address: The redis server address.
        select: The redis database for mail data.
        prefix: The prefix for mail data keys.
        users_hash: The hash key for user lookup.
        users_key: The user lookup key template.
        users_json: True if the user lookup value contains JSON.

    """

    def __init__(self, args: Namespace, *, address: str,
                 select: Optional[int], prefix: str,
                 users_hash: Optional[str], users_key: str,
                 users_json: bool, **extra: Any) -> None:
        super().__init__(args, **extra)
        self._address = address
        self._select = select
        self._prefix = prefix
        self._users_hash = users_hash
        self._users_key = users_key
        self._users_json = users_json

    @property
    def address(self) -> str:
        """The redis server address. Defaults to a connection to localhost.

        See Also:
            :func:`aioredis.create_connection`

        """
        return self._address

    @property
    def select(self) -> Optional[int]:
        """The redis database for mail data. If given, the `SELECT`_ command is
        called after successful user lookup.

        .. _SELECT: https://redis.io/commands/select

        """
        return self._select

    @property
    def prefix(self) -> str:
        """The prefix for mail data keys. This prefix does not apply to user
        lookup, e.g. :attr:`.users_hash` or :attr:`.users_key`.

        """
        return self._prefix

    @property
    def users_hash(self) -> Optional[str]:
        """The name of the hash where keys are looked up for user login. If
        this value is None, which is the default, keys are looked up in the
        full redis database.

        """
        return self._users_hash

    @property
    def users_key(self) -> str:
        """The key template for looking up user login. The login name
        (:attr:`~pysasl.AuthenticationCredentials.authcid`) is substituted in
        the template in place of ``{name}`` using :meth:`str.format`. By
        default, the template is ``{name}`` so the key will be the unaltered
        login name.

        """
        return self._users_key

    @property
    def users_json(self) -> bool:
        """True if the value from the user lookup key contains a JSON object
        with a ``"password"`` attribute, instead of simply containing the
        password value encoded with UTF-8.

        """
        return self._users_json

    @property
    def prefix_override_hash(self) -> str:
        """The name of a hash used to override user prefixes -- not needed for
        normal operation.

        Typically, the keys used for user data after login are prefixed with
        :attr:`.prefix` and the SHA-1 hash of the login name.  However, if a
        login name is changed, it can still use to its old data by setting
        :attr:`.users_key` in this hash to the prefix before the name change.

        """
        return f'{self.prefix}_prefix'

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {'address': args.address,
                'select': args.select,
                'prefix': args.prefix,
                'users_hash': args.users_hash,
                'users_key': args.users_key,
                'users_json': args.users_json}


class Session(BaseSession[Message]):
    """The session implementation for the redis backend."""

    resource = __name__

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> IMAPConfig:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set

    @classmethod
    async def login(cls, credentials: AuthenticationCredentials,
                    config: Config) -> 'Session':
        """Checks the given credentials for a valid login and returns a new
        session.

        """
        redis = await create_redis(config.address)
        prefix = await cls._check_user(redis, config, credentials)
        if config.select is not None:
            await redis.select(config.select)
        mailbox_set = MailboxSet(redis, prefix)
        try:
            await mailbox_set.add_mailbox('INBOX')
        except MailboxConflict:
            pass
        filter_set = FilterSet(redis, prefix)
        return cls(credentials.identity, config, mailbox_set, filter_set)

    @classmethod
    async def _check_user(cls, redis: Redis, config: Config,
                          credentials: AuthenticationCredentials) -> bytes:
        user = credentials.authcid
        password, prefix = await cls._get_password(redis, config, user)
        if user != credentials.identity:
            raise InvalidAuth()
        elif ldap_context is None or not credentials.has_secret:
            if not credentials.check_secret(password):
                raise InvalidAuth()
        elif not ldap_context.verify(credentials.secret, password):
            raise InvalidAuth()
        return prefix

    @classmethod
    async def _get_password(cls, redis: Redis, config: Config,
                            name: str) -> Tuple[str, bytes]:
        key = config.users_key.format(name=name)
        multi = redis.multi_exec()
        if config.users_hash is None:
            multi.get(key)
        else:
            multi.hget(config.users_hash, key)
        multi.hget(config.prefix_override_hash, key)
        value, prefix = await multi.execute()
        if prefix is None:
            prefix = cls._get_prefix(name, config.prefix)
        if value is None:
            raise InvalidAuth()
        elif config.users_json:
            value_obj = json.loads(value)
            try:
                return value_obj['password'], prefix
            except KeyError as exc:
                raise InvalidAuth() from exc
        else:
            return value.decode('utf-8'), prefix

    @classmethod
    def _get_prefix(cls, name: str, root_prefix: str) -> bytes:
        name_bytes = name.encode('utf-8')
        hash_obj = hashlib.sha1()
        hash_obj.update(name_bytes)
        prefix = root_prefix + hash_obj.hexdigest()
        return prefix.encode('utf-8')



from datetime import datetime, tzinfo
from typing import Tuple, Optional

from .. import Params, Parseable
from ..exceptions import InvalidContent
from ..primitives import QuotedString
from ...bytes import BytesFormat

__all__ = ['DateTime']


class DateTime(Parseable[datetime]):
    """Represents a date-time quoted string from an IMAP stream.

    Args:
        when: The date-time value.
        raw: The raw bytestring from IMAP parsing.

    """

    def __init__(self, when: datetime, raw: bytes = None) -> None:
        super().__init__()
        if when.tzinfo is None:
            when = when.replace(tzinfo=self.get_local_tzinfo())
        self.when = when
        self._raw = raw

    @property
    def value(self) -> datetime:
        """The date-time value."""
        return self.when

    @classmethod
    def get_local_tzinfo(cls) -> Optional[tzinfo]:
        """The system timezone, used when no timezone is specified."""
        return datetime.now().astimezone().tzinfo

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple['DateTime', memoryview]:
        string, after = QuotedString.parse(buf, params)
        try:
            when_str = str(string.value, 'ascii')
            when = datetime.strptime(when_str, '%d-%b-%Y %X %z')
        except (UnicodeDecodeError, ValueError):
            raise InvalidContent(buf)
        return cls(when, string.value), after

    def __bytes__(self) -> bytes:
        if self._raw is None:
            raw_str = self.value.strftime('%d-%b-%Y %X %z')
            self._raw = bytes(raw_str, 'ascii')
        return BytesFormat(b'"%b"') % (self._raw, )



"""Package defining all the IMAP parsing and response classes."""

from typing import Optional

from .response import ResponseCode
from ..bytes import MaybeBytes

__all__ = ['RequiresContinuation', 'NotParseable', 'InvalidContent',
           'UnexpectedType']


class RequiresContinuation(Exception):
    """Indicates that the buffer has been successfully parsed so far, but
    requires a continuation of the command from the client.

    Args:
        message: The message from the server.
        literal_length: If the continuation is for a string literal, this is
            the byte length to expect.

    """

    __slots__ = ['message', 'literal_length']

    def __init__(self, message: bytes, literal_length: int = 0) -> None:
        super().__init__()
        self.message = message
        self.literal_length = literal_length


class NotParseable(Exception):
    """Indicates that the given buffer was not parseable by one or all of the
    data formats.

    Args:
        buf: The buffer with the parsing error.

    """

    error_indicator = b'[:ERROR:]'

    __slots__ = ['buf', 'code', '_raw']

    def __init__(self, buf: memoryview, code:
                 Optional[MaybeBytes] = None) -> None:
        super().__init__()
        self.buf = bytes(buf)
        self.code = ResponseCode.of(code)
        self._raw: Optional[bytes] = None

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return self._raw
        self._raw = raw = self.error_indicator.join((b'', self.buf))
        return raw

    def __str__(self) -> str:
        return str(bytes(self), 'ascii', 'replace')


class InvalidContent(NotParseable, ValueError):
    """Indicates the type of the parsed content was correct, but something
    about the content did not fit what was expected by the special type.

    """
    pass


class UnexpectedType(NotParseable):
    """Indicates that a generic parseable that was given a sub-type expectation
    failed to meet that expectation.

    """
    pass




import asyncio
import uuid
from typing import Optional, Tuple, Sequence, List

from aioredis import Redis, MultiExecError, WatchVariableError  # type: ignore

from pymap.filter import EntryPointFilterSet

__all__ = ['FilterSet']


async def _check_errors(multi) -> bool:
    # Prevents warning about exception never being retrieved.
    errors = await asyncio.gather(*multi._results, return_exceptions=True)
    return any(not isinstance(exc, WatchVariableError) for exc in errors)


class FilterSet(EntryPointFilterSet[bytes]):

    _active_name = b''

    def __init__(self, redis: Redis, prefix: bytes) -> None:
        super().__init__('sieve')
        self._redis = redis
        self._prefix = prefix

    @classmethod
    def _get_random_key(cls) -> bytes:
        return uuid.uuid4().hex.encode('ascii')

    async def put(self, name: str, value: bytes) -> None:
        redis = self._redis
        prefix = self._prefix
        name_bytes = name.encode('utf-8')
        key = self._get_random_key()
        await redis.hset(prefix + b':sieve-data', key, value)
        await redis.hset(prefix + b':sieve', name_bytes, key)

    async def delete(self, name: str) -> None:
        redis = self._redis
        prefix = self._prefix
        name_bytes = name.encode('utf-8')
        while True:
            await redis.watch(prefix + b':sieve')
            name_key, active_key = await redis.hmget(
                prefix + b':sieve', name_bytes, self._active_name)
            if name_key is None:
                await redis.unwatch()
                raise KeyError(name)
            elif name_key == active_key:
                await redis.unwatch()
                raise ValueError(name)
            multi = redis.multi_exec()
            multi.hdel(prefix + b':sieve-data', name_key)
            multi.hdel(prefix + b':sieve', name_bytes)
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break

    async def rename(self, before_name: str, after_name: str) -> None:
        redis = self._redis
        prefix = self._prefix
        before_name_bytes = before_name.encode('utf-8')
        after_name_bytes = after_name.encode('utf-8')
        while True:
            await redis.watch(prefix + b':sieve')
            before_name_key, after_name_key = await redis.hmget(
                prefix + b':sieve', before_name_bytes, after_name_bytes)
            if before_name_key is None:
                await redis.unwatch()
                raise KeyError(before_name)
            elif after_name_key is not None:
                await redis.unwatch()
                raise KeyError(after_name)
            multi = redis.multi_exec()
            multi.hdel(prefix + b':sieve', before_name_bytes)
            multi.hset(prefix + b':sieve', after_name_bytes, before_name_key)
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break

    async def set_active(self, name: Optional[str]) -> None:
        redis = self._redis
        prefix = self._prefix
        if name is None:
            await redis.hdel(prefix + b':sieve', self._active_name)
            return
        else:
            name_bytes = name.encode('utf-8')
            while True:
                await redis.watch(prefix + b':sieve')
                name_key = await redis.hget(prefix + b':sieve', name_bytes)
                if name_key is None:
                    await redis.unwatch()
                    raise KeyError(name)
                multi = redis.multi_exec()
                multi.hset(prefix + b':sieve', self._active_name, name_key)
                try:
                    await multi.execute()
                except MultiExecError:
                    if await _check_errors(multi):
                        raise
                else:
                    break

    async def get(self, name: str) -> bytes:
        redis = self._redis
        prefix = self._prefix
        name_bytes = name.encode('utf-8')
        name_key = await redis.hget(prefix + b':sieve', name_bytes)
        if name_key is None:
            raise KeyError(name)
        value = await redis.hget(prefix + b':sieve-data', name_key)
        if value is None:
            raise KeyError(name)
        return value

    async def get_active(self) -> Optional[bytes]:
        redis = self._redis
        prefix = self._prefix
        active_key = await redis.hget(prefix + b':sieve', self._active_name)
        if active_key is None:
            return None
        value = await redis.hget(prefix + b':sieve-data', active_key)
        if value is None:
            return None
        return value

    async def get_all(self) -> Tuple[Optional[str], Sequence[str]]:
        redis = self._redis
        prefix = self._prefix
        sieve_names = await redis.hgetall(prefix + b':sieve')
        active_key: Optional[bytes] = sieve_names.get(self._active_name)
        active_name: Optional[str] = None
        names: List[str] = []
        for name, key in sieve_names.items():
            if name != self._active_name:
                name_str = name.decode('utf-8')
                names.append(name_str)
                if key == active_key:
                    active_name = name_str
        return active_name, names




from abc import abstractmethod
from asyncio import shield
from typing import Generic, Tuple, Optional, FrozenSet, Iterable, \
    Sequence, List

from pymap.concurrent import Event
from pymap.config import IMAPConfig
from pymap.exceptions import MailboxNotFound, MailboxReadOnly
from pymap.flags import FlagOp, SessionFlags, PermanentFlags
from pymap.interfaces.filter import FilterSetInterface
from pymap.interfaces.message import AppendMessage
from pymap.interfaces.session import SessionInterface
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.parsing.specials import SequenceSet, SearchKey, \
    FetchAttribute, FetchRequirement
from pymap.parsing.specials.flag import Flag, Seen
from pymap.parsing.response.code import AppendUid, CopyUid
from pymap.search import SearchParams, SearchCriteriaSet
from pymap.selected import SelectedMailbox

from .mailbox import MailboxDataInterface, MailboxSetInterface, MessageT

__all__ = ['BaseSession']


class BaseSession(SessionInterface, Generic[MessageT]):
    """Base implementation of
    :class:`~pymap.interfaces.session.SessionInterface` intended for use by
    most backends.

    Args:
        owner: The logged-in user name.

    """

    def __init__(self, owner: str) -> None:
        super().__init__()
        self._owner = owner

    @property
    def owner(self) -> str:
        return self._owner

    @property
    @abstractmethod
    def config(self) -> IMAPConfig:
        ...

    @property
    @abstractmethod
    def mailbox_set(self) \
            -> MailboxSetInterface[MailboxDataInterface[MessageT]]:
        ...

    @property
    def filter_set(self) -> Optional[FilterSetInterface]:
        return None

    async def _load_updates(self, selected: Optional[SelectedMailbox],
                            mbx: Optional[MailboxDataInterface[MessageT]]) \
            -> Optional[SelectedMailbox]:
        if selected:
            if not mbx or selected.name != mbx.name:
                try:
                    mbx = await self.mailbox_set.get_mailbox(selected.name)
                except MailboxNotFound:
                    selected.set_deleted()
                    return selected
            return await mbx.update_selected(selected)
        return selected

    @classmethod
    def _find_selected(cls, selected: Optional[SelectedMailbox],
                       mbx: MailboxDataInterface[MessageT]) \
            -> Optional[SelectedMailbox]:
        if selected and selected.name == mbx.name:
            return selected
        return mbx.selected_set.any_selected

    async def list_mailboxes(self, ref_name: str, filter_: str,
                             subscribed: bool = False,
                             selected: SelectedMailbox = None) \
            -> Tuple[Iterable[Tuple[str, Optional[str], Sequence[bytes]]],
                     Optional[SelectedMailbox]]:
        delimiter = self.mailbox_set.delimiter
        if filter_:
            list_tree = ListTree(delimiter).update('INBOX')
            if subscribed:
                list_tree.update(*await self.mailbox_set.list_subscribed())
            else:
                list_tree.update(*await self.mailbox_set.list_mailboxes())
            ret = [(entry.name, delimiter, entry.attributes)
                   for entry in list_tree.list_matching(ref_name, filter_)]
        else:
            ret = [("", delimiter, [b'Noselect'])]
        return ret, await self._load_updates(selected, None)

    async def get_mailbox(self, name: str, selected: SelectedMailbox = None) \
            -> Tuple[MailboxSnapshot, Optional[SelectedMailbox]]:
        mbx = await self.mailbox_set.get_mailbox(name)
        snapshot = await mbx.snapshot()
        return snapshot, await self._load_updates(selected, mbx)

    async def create_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        await self.mailbox_set.add_mailbox(name)
        return await self._load_updates(selected, None)

    async def delete_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        await self.mailbox_set.delete_mailbox(name)
        return await self._load_updates(selected, None)

    async def rename_mailbox(self, before_name: str, after_name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        await self.mailbox_set.rename_mailbox(before_name, after_name)
        return await self._load_updates(selected, None)

    async def subscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        mbx = await self.mailbox_set.get_mailbox('INBOX')
        await self.mailbox_set.set_subscribed(name, True)
        return await self._load_updates(selected, mbx)

    async def unsubscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        mbx = await self.mailbox_set.get_mailbox('INBOX')
        await self.mailbox_set.set_subscribed(name, False)
        return await self._load_updates(selected, mbx)

    async def append_messages(self, name: str,
                              messages: Sequence[AppendMessage],
                              selected: SelectedMailbox = None) \
            -> Tuple[AppendUid, Optional[SelectedMailbox]]:
        mbx = await self.mailbox_set.get_mailbox(name, try_create=True)
        if mbx.readonly:
            raise MailboxReadOnly(name)
        dest_selected = self._find_selected(selected, mbx)
        uids: List[int] = []
        for append_msg in messages:
            msg = await mbx.add(append_msg, recent=not dest_selected)
            if dest_selected:
                dest_selected.session_flags.add_recent(msg.uid)
            uids.append(msg.uid)
        mbx.selected_set.updated.set()
        return (AppendUid(mbx.uid_validity, uids),
                await self._load_updates(selected, mbx))

    async def select_mailbox(self, name: str, readonly: bool = False) \
            -> Tuple[MailboxSnapshot, SelectedMailbox]:
        mbx = await self.mailbox_set.get_mailbox(name)
        selected = SelectedMailbox(name, readonly or mbx.readonly,
                                   PermanentFlags(mbx.permanent_flags),
                                   SessionFlags(mbx.session_flags),
                                   selected_set=mbx.selected_set)
        if not selected.readonly:
            await mbx.claim_recent(selected)
        snapshot = await mbx.snapshot()
        return snapshot, await mbx.update_selected(selected)

    async def check_mailbox(self, selected: SelectedMailbox, *,
                            wait_on: Event = None,
                            housekeeping: bool = False) -> SelectedMailbox:
        mbx = await self.mailbox_set.get_mailbox(selected.name)
        if housekeeping:
            await shield(mbx.cleanup())
        if wait_on is not None:
            either_event = wait_on.or_event(mbx.selected_set.updated)
            await either_event.wait()
        return await mbx.update_selected(selected)

    async def fetch_messages(self, selected: SelectedMailbox,
                             sequence_set: SequenceSet,
                             attributes: FrozenSet[FetchAttribute]) \
            -> Tuple[Iterable[Tuple[int, MessageT]], SelectedMailbox]:
        mbx = await self.mailbox_set.get_mailbox(selected.name)
        req = FetchRequirement.reduce({attr.requirement
                                       for attr in attributes})
        ret = [(seq, msg) async for seq, msg
               in mbx.find(sequence_set, selected, req)]
        if not selected.readonly and any(attr.set_seen for attr in attributes):
            seen_set = frozenset([Seen])
            await mbx.update_flags([msg for _, msg in ret],
                                   seen_set, FlagOp.ADD)
            mbx.selected_set.updated.set()
        return ret, await mbx.update_selected(selected)

    async def search_mailbox(self, selected: SelectedMailbox,
                             keys: FrozenSet[SearchKey]) \
            -> Tuple[Iterable[Tuple[int, MessageT]], SelectedMailbox]:
        mbx = await self.mailbox_set.get_mailbox(selected.name)
        req = FetchRequirement.reduce({key.requirement for key in keys})
        ret: List[Tuple[int, MessageT]] = []
        params = SearchParams(selected,
                              disabled=self.config.disable_search_keys)
        search = SearchCriteriaSet(keys, params)
        async for seq, msg in mbx.find(search.sequence_set, selected, req):
            if search.matches(seq, msg):
                ret.append((seq, msg))
        return ret, await mbx.update_selected(selected)

    async def expunge_mailbox(self, selected: SelectedMailbox,
                              uid_set: SequenceSet = None) -> SelectedMailbox:
        if selected.readonly:
            raise MailboxReadOnly(selected.name)
        mbx = await self.mailbox_set.get_mailbox(selected.name)
        if not uid_set:
            uid_set = SequenceSet.all(uid=True)
        expunge_uids = await mbx.find_deleted(uid_set, selected)
        await mbx.delete(expunge_uids)
        mbx.selected_set.updated.set()
        return await mbx.update_selected(selected)

    async def copy_messages(self, selected: SelectedMailbox,
                            sequence_set: SequenceSet,
                            mailbox: str) \
            -> Tuple[Optional[CopyUid], SelectedMailbox]:
        mbx = await self.mailbox_set.get_mailbox(selected.name)
        dest = await self.mailbox_set.get_mailbox(mailbox, try_create=True)
        if dest.readonly:
            raise MailboxReadOnly(mailbox)
        req = FetchRequirement.BODY
        dest_selected = self._find_selected(selected, dest)
        uids: List[Tuple[int, int]] = []
        async for _, msg in mbx.find(sequence_set, selected, req):
            if not msg.expunged:
                source_uid = msg.uid
                msg = await dest.add(msg.append_msg, recent=not dest_selected)
                if dest_selected:
                    dest_selected.session_flags.add_recent(msg.uid)
                uids.append((source_uid, msg.uid))
        dest.selected_set.updated.set()
        return (CopyUid(dest.uid_validity, uids),
                await mbx.update_selected(selected))

    async def update_flags(self, selected: SelectedMailbox,
                           sequence_set: SequenceSet,
                           flag_set: FrozenSet[Flag],
                           mode: FlagOp = FlagOp.REPLACE) \
            -> Tuple[Iterable[Tuple[int, MessageT]], SelectedMailbox]:
        if selected.readonly:
            raise MailboxReadOnly(selected.name)
        mbx = await self.mailbox_set.get_mailbox(selected.name)
        permanent_flags = selected.permanent_flags & flag_set
        messages: List[Tuple[int, MessageT]] = []
        async for msg_seq, msg in mbx.find(sequence_set, selected):
            if not msg.expunged:
                selected.session_flags.update(msg.uid, flag_set, mode)
            messages.append((msg_seq, msg))
        await mbx.update_flags([msg for _, msg in messages],
                               permanent_flags, mode)
        mbx.selected_set.updated.set()
        return messages, await mbx.update_selected(selected)



"""Append a message directly to a user's mailbox."""

import sys
import time
import traceback
from argparse import ArgumentParser, FileType
from typing import Type, Tuple, TextIO

from .command import ClientCommand
from ..grpc.admin_pb2 import AppendRequest, AppendResponse, \
    SUCCESS, ERROR_RESPONSE


class AppendCommand(ClientCommand):

    success = '2.0.0 Message delivered'
    messages = {'InvalidAuth': '5.7.8 Authentication credentials invalid',
                'TimedOut': '4.4.2 Connection timed out',
                'ConnectionFailed': '4.3.0 Connection failed',
                'UnhandledError': '4.3.0 Unhandled system error',
                'MailboxNotFound': '4.2.0 Message not deliverable',
                'AppendFailure': '4.2.0 Message not deliverable'}

    @classmethod
    def init(cls, parser: ArgumentParser, subparsers) \
            -> Tuple[str, Type['AppendCommand']]:
        subparser = subparsers.add_parser(
            'append', description=__doc__,
            help='append a message to a mailbox')
        subparser.add_argument('--from', metavar='ADDRESS', dest='sender',
                               default='', help='the message envelope sender')
        subparser.add_argument('--to', metavar='ADDRESS', dest='recipient',
                               help='the message envelope recipient')
        subparser.add_argument('--mailbox', metavar='NAME',
                               help='the mailbox name')
        subparser.add_argument('--timestamp', type=int, metavar='SECONDS',
                               help='the message timestamp (defualt: now)')
        subparser.add_argument('--data', type=FileType('rb'), metavar='FILE',
                               default=sys.stdin.buffer,
                               help='the message data (default: stdin)')
        subparser.add_argument('user', help='the user name')
        flags = subparser.add_argument_group('message flags')
        flags.add_argument('--flag', dest='flags', action='append',
                           metavar='VAL', help='a message flag or keyword')
        flags.add_argument('--flagged', dest='flags', action='append_const',
                           const='\\Flagged', help='the message is flagged')
        flags.add_argument('--seen', dest='flags', action='append_const',
                           const='\\Seen', help='the message is seen')
        flags.add_argument('--draft', dest='flags', action='append_const',
                           const='\\Draft', help='the message is a draft')
        flags.add_argument('--deleted', dest='flags', action='append_const',
                           const='\\Deleted', help='the message is deleted')
        flags.add_argument('--answered', dest='flags', action='append_const',
                           const='\\Answered', help='the message is answered')
        return 'append', cls

    async def run(self, fileobj: TextIO) -> int:
        args = self.args
        recipient = args.recipient or args.user
        data = args.data.read()
        when: int = args.timestamp or int(time.time())
        req = AppendRequest(user=args.user, sender=args.sender,
                            recipient=recipient, mailbox=args.mailbox,
                            data=data, flags=args.flags, when=when)
        try:
            res = await self.stub.Append(req)
        except OSError:
            traceback.print_exc()
            code, msg = 1, self.messages['ConnectionFailed']
        except Exception:
            traceback.print_exc()
            code, msg = 1, self.messages['UnhandledError']
        else:
            print(res, file=sys.stderr)
            code, msg = self._parse(res)
        print(msg, file=fileobj)
        return code

    def _parse(self, response: AppendResponse) -> Tuple[int, str]:
        if response.result == SUCCESS:
            return 0, self.success
        elif response.result == ERROR_RESPONSE:
            try:
                msg = self.messages[response.error_type]
            except KeyError:
                msg = self.messages['UnhandledError']
            return 1, msg
        else:
            raise NotImplementedError(response.result)




import os.path
from argparse import Namespace, ArgumentDefaultsHelpFormatter
from concurrent.futures import ThreadPoolExecutor
from typing import TypeVar, Any, Type, Optional, Tuple, Mapping

from pysasl import AuthenticationCredentials

from pymap.concurrent import Subsystem
from pymap.config import IMAPConfig
from pymap.exceptions import InvalidAuth
from pymap.filter import EntryPointFilterSet, SingleFilterSet
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.session import LoginProtocol

from .layout import MaildirLayout
from .mailbox import Message, Maildir, MailboxSet
from ..session import BaseSession

__all__ = ['MaildirBackend', 'Config', 'Session']

_SessionT = TypeVar('_SessionT', bound='Session')


class MaildirBackend(BackendInterface):
    """Defines an on-disk backend that uses :class:`~mailbox.Maildir` for
    mailbox storage and `MailboxFormat/Maildir
    <https://wiki2.dovecot.org/MailboxFormat/Maildir>`_ for metadata storage.

    """

    def __init__(self, login: LoginProtocol, config: IMAPConfig) -> None:
        super().__init__()
        self._login = login
        self._config = config

    @property
    def login(self) -> LoginProtocol:
        return self._login

    @property
    def config(self) -> IMAPConfig:
        return self._config

    @classmethod
    def add_subparser(cls, subparsers) -> None:
        parser = subparsers.add_parser(
            'maildir', help='on-disk backend',
            formatter_class=ArgumentDefaultsHelpFormatter)
        parser.add_argument('users_file', help='path the the users file')
        parser.add_argument('--base-dir', metavar='DIR',
                            help='base directory for mailbox relative paths')
        parser.add_argument('--concurrency', metavar='NUM', type=int,
                            help='maximum number of IO workers')
        parser.add_argument('--layout', metavar='TYPE', default='++',
                            help='maildir directory layout')

    @classmethod
    async def init(cls, args: Namespace) -> Tuple['MaildirBackend', 'Config']:
        config = Config.from_args(args)
        return cls(Session.login, config), config


class Config(IMAPConfig):
    """The config implementation for the maildir backend.

    Args:
        args: The command-line arguments.
        users_file: The path to the users file.
        base_dir: The base directory for all relative mailbox paths.
        layout: The Maildir directory layout.

    """

    def __init__(self, args: Namespace, *, users_file: str,
                 base_dir: Optional[str], layout: str, **extra: Any) -> None:
        super().__init__(args, **extra)
        self._users_file = users_file
        self._base_dir = self._get_base_dir(base_dir, users_file)
        self._layout = layout

    @classmethod
    def _get_base_dir(cls, base_dir: Optional[str],
                      users_file: Optional[str]) -> str:
        if base_dir:
            return base_dir
        elif users_file:
            return os.path.dirname(users_file)
        else:
            raise ValueError('--base-dir', base_dir)

    @property
    def users_file(self) -> str:
        """Used by the default :meth:`~Session.find_user` implementation
        to retrieve the users file path from the command-line arguments. The
        users file is given as the first positional argument on the
        command-line.

        This file contains a valid login on each line, which are split into
        three parts by colon (``:``) characters: the user name, the mailbox
        path, and the password.

        The password may contain colon characters. The mailbox path may be
        empty, relative, or absolute. If it is empty, the user ID is used as a
        relative path.

        """
        return self._users_file

    @property
    def base_dir(self) -> str:
        """The base directory for all relative mailbox paths. The default is
        the directory containing the users file.

        """
        return self._base_dir

    @property
    def layout(self) -> str:
        """The Maildir directory layout name.

        See Also:
            :class:`~pymap.backend.maildir.layout.MaildirLayout`

        """
        return self._layout

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        executor = ThreadPoolExecutor(args.concurrency)
        subsystem = Subsystem.for_executor(executor)
        return {'users_file': args.users_file,
                'base_dir': args.base_dir,
                'layout': args.layout,
                'subsystem': subsystem}


class FilterSet(EntryPointFilterSet[bytes], SingleFilterSet[bytes]):

    def __init__(self, user_dir: str) -> None:
        super().__init__('sieve')
        self._user_dir = user_dir

    async def replace_active(self, value: Optional[bytes]) -> None:
        path = os.path.join(self._user_dir, 'dovecot.sieve')
        if value is None:
            try:
                os.unlink(path)
            except FileNotFoundError:
                pass
        else:
            with open(path, 'wb') as sieve_file:
                sieve_file.write(value)

    async def get_active(self) -> Optional[bytes]:
        path = os.path.join(self._user_dir, 'dovecot.sieve')
        try:
            with open(path, 'rb') as sieve_file:
                return sieve_file.read()
        except FileNotFoundError:
            return None


class Session(BaseSession[Message]):
    """The session implementation for the maildir backend."""

    resource = __name__

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> Config:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set

    @classmethod
    async def login(cls: Type[_SessionT],
                    credentials: AuthenticationCredentials,
                    config: Config) -> _SessionT:
        """Checks the given credentials for a valid login and returns a new
        session.

        """
        user = credentials.authcid
        password, user_dir = await cls.find_user(config, user)
        if user != credentials.identity:
            raise InvalidAuth()
        elif not credentials.check_secret(password):
            raise InvalidAuth()
        maildir, layout = cls._load_maildir(config, user_dir)
        mailbox_set = MailboxSet(maildir, layout)
        filter_set = FilterSet(layout.path)
        return cls(credentials.identity, config, mailbox_set, filter_set)

    @classmethod
    async def find_user(cls, config: Config, user: str) \
            -> Tuple[str, str]:
        """If the given user ID exists, return its expected password and
        mailbox path. Override this method to implement custom login logic.

        Args:
            config: The maildir config object.
            user: The expected user ID.

        Raises:
            InvalidAuth: The user ID was not valid.

        """
        with open(config.users_file, 'r') as users_file:
            for line in users_file:
                this_user, user_dir, password = line.split(':', 2)
                if user == this_user:
                    return password.rstrip('\r\n'), user_dir or user
        raise InvalidAuth()

    @classmethod
    def _load_maildir(cls, config: Config, user_dir: str) \
            -> Tuple[Maildir, MaildirLayout]:
        full_path = os.path.join(config.base_dir, user_dir)
        layout = MaildirLayout.get(full_path, config.layout, Maildir)
        create = not os.path.exists(full_path)
        return Maildir(full_path, create=create), layout



"""Base implementation of the :mod:`pymap.interfaces.mailbox` interfaces."""

import random
import time
from typing import Optional, Iterable, FrozenSet
from typing_extensions import Final

from .interfaces.mailbox import MailboxInterface
from .parsing.specials.flag import Flag, Recent

__all__ = ['MailboxSnapshot']


class MailboxSnapshot(MailboxInterface):
    """Implements the full functionality of a mailbox using entirely static
    arguments to the constructor. Backends can create and return a new
    mailbox snapshot every time a
    :class:`~pymap.interfaces.mailbox.MailboxInterface` is required.

    Args:
        name: The name of the mailbox.
        readonly: If ``True``, the mailbox is read-only.
        uid_validity: The UID validity value for mailbox consistency.
        permanent_flags: The permanent flags defined in the mailbox.
        session_flags: The session flags defined in the mailbox.
        exists: Number of total messages in the mailbox.
        recent: Number of recent messages in the mailbox.
        unseen: Number of unseen messages in the mailbox.
        first_unseen: The sequence number of the first unseen message.
        next_uid: The predicted next message UID.

    """

    __slots__ = ['name', 'readonly', 'uid_validity', 'permanent_flags',
                 'session_flags', 'exists', 'recent', 'unseen', 'first_unseen',
                 'next_uid']

    def __init__(self, name: str, readonly: bool, uid_validity: int,
                 permanent_flags: Iterable[Flag],
                 session_flags: FrozenSet[Flag],
                 exists: int, recent: int, unseen: int,
                 first_unseen: Optional[int], next_uid: int) -> None:
        super().__init__()
        self.name: Final = name
        self.readonly: Final = readonly
        self.uid_validity: Final = uid_validity
        self.permanent_flags: Final = frozenset(permanent_flags) - {Recent}
        self.session_flags: Final = frozenset(session_flags) | {Recent}
        self.exists: Final = exists
        self.recent: Final = recent
        self.unseen: Final = unseen
        self.first_unseen: Final = first_unseen
        self.next_uid: Final = next_uid

    @classmethod
    def new_uid_validity(cls) -> int:
        """Generate a new UID validity value for a mailbox, where the first
        two bytes are time-based and the second two bytes are random.

        """
        time_part = int(time.time()) % 4096
        rand_part = random.randint(0, 1048576)
        return (time_part << 20) + rand_part

    @property
    def flags(self) -> FrozenSet[Flag]:
        return self.permanent_flags | self.session_flags




from typing import Tuple, Optional

from . import AString
from .. import Params, Parseable
from ..modutf7 import modutf7_encode, modutf7_decode

__all__ = ['Mailbox']


class Mailbox(Parseable[str]):
    """Represents a mailbox data object from an IMAP stream.

    Args:
        mailbox: The mailbox name.

    """

    def __init__(self, mailbox: str) -> None:
        super().__init__()
        if mailbox.upper() == 'INBOX':
            self.mailbox = 'INBOX'
            self._raw: Optional[bytes] = b'INBOX'
        else:
            self.mailbox = mailbox
            self._raw = None

    @property
    def value(self) -> str:
        """The mailbox name."""
        return self.mailbox

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple['Mailbox', memoryview]:
        atom, buf = AString.parse(buf, params)
        mailbox = atom.value
        if mailbox.upper() == b'INBOX':
            return cls('INBOX'), buf
        return cls(modutf7_decode(mailbox)), buf

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return self._raw
        self._raw = raw = bytes(AString(modutf7_encode(self.value)))
        return raw

    def __str__(self) -> str:
        return self.value



"""Defines useful types and utilities for working with bytestrings."""

import re
from abc import abstractmethod, ABCMeta
from itertools import chain
from typing import cast, Any, TypeVar, ByteString, SupportsBytes, Union, \
    Callable, Iterable, Optional, Pattern, Match
from typing_extensions import Protocol

__all__ = ['MaybeBytes', 'MaybeBytesT', 'WriteStream', 'Writeable',
           'BytesFormat', 'ReView', 'rev']

#: A bytes object, memoryview,  or an object with a ``__bytes__`` method.
MaybeBytes = Union[ByteString, memoryview, SupportsBytes]

#: A type variable bound to :class:`MaybeBytes`.
MaybeBytesT = TypeVar('MaybeBytesT', bound=MaybeBytes)

_FormatArg = Union[MaybeBytes, int]


class WriteStream(Protocol):
    """Typing protocol indicating the object implements the :meth:`.write`
    method.

    See Also:
        :class:`~asyncio.StreamWriter`, :class:`~typing.BinaryIO`

    """

    @abstractmethod
    def write(self, data: bytes) -> Any:
        """Defines an abstract method where ``data`` is written to a stream or
        buffer.

        Args:
            data: The data to write.

        """
        ...


class Writeable(metaclass=ABCMeta):
    """Base class for types that can be written to a stream."""

    @classmethod
    def empty(cls) -> 'Writeable':
        """Return a :class:`Writeable` for an empty string."""
        return _EmptyWriteable()

    @classmethod
    def wrap(cls, data: MaybeBytes) -> 'Writeable':
        """Wrap the bytes in a :class:`Writeable`.

        Args:
            data: The object to wrap.

        """
        return _WrappedWriteable(data)

    @classmethod
    def concat(cls, data: Iterable[MaybeBytes]) -> 'Writeable':
        """Wrap the iterable in a :class:`Writeable` that will write eachitem.

        Args:
            data: The iterable to wrap.

        """
        return _ConcatWriteable(data)

    def write(self, writer: WriteStream) -> None:
        """Write the object to the stream, with one or more calls to
        :meth:`~WriteStream.write`.

        Args:
            writer: The output stream.

        """
        writer.write(bytes(self))

    def __len__(self) -> int:
        return len(bytes(self))

    @abstractmethod
    def __bytes__(self) -> bytes:
        ...


class _EmptyWriteable(Writeable):

    def write(self, writer: WriteStream) -> None:
        pass

    def __bytes__(self) -> bytes:
        return b''

    def __repr__(self) -> str:
        return '<Writeable empty>'


class _WrappedWriteable(Writeable):

    __slots__ = ['data']

    def __init__(self, data: MaybeBytes) -> None:
        if isinstance(data, bytes):
            self.data = data
        else:
            self.data = bytes(data)

    def __bytes__(self) -> bytes:
        return self.data

    def __repr__(self) -> str:
        return f'<Writeable {repr(self.data)}>'


class _ConcatWriteable(Writeable):

    __slots__ = ['data']

    def __init__(self, data: Iterable[MaybeBytes]) -> None:
        self.data = list(data)

    def write(self, writer: WriteStream) -> None:
        for item in self.data:
            if isinstance(item, Writeable):
                item.write(writer)
            else:
                writer.write(bytes(item))

    def __bytes__(self) -> bytes:
        return BytesFormat(b'').join(self.data)

    def __repr__(self) -> str:
        return f'<Writeable {repr(self.data)}>'


class BytesFormat:
    """Helper utility for performing formatting operations that produce
    bytestrings. While similar to the builtin formatting and join
    operations, this class intends to provide cleaner typing.

    Args:
        how: The formatting string or join delimiter to use.

    """

    __slots__ = ['how']

    def __init__(self, how: bytes) -> None:
        super().__init__()
        self.how = how

    def __mod__(self, other: Union[_FormatArg, Iterable[_FormatArg]]) -> bytes:
        """String interpolation, shortcut for :meth:`.format`.

        Args:
            other: The data interpolated into the format string.

        """
        if isinstance(other, bytes):
            return self.format([other])
        elif hasattr(other, '__bytes__'):
            supports_bytes = cast(SupportsBytes, other)
            return self.format([bytes(supports_bytes)])
        elif hasattr(other, '__iter__'):
            items = cast(Iterable[_FormatArg], other)
            return self.format(items)
        return NotImplemented

    @classmethod
    def _fix_format_arg(cls, data: _FormatArg) -> Any:
        if isinstance(data, int):
            return data
        else:
            return bytes(data)

    def format(self, data: Iterable[_FormatArg]) -> bytes:
        """String interpolation into the format string.

        Args:
            data: The data interpolated into the format string.

        Examples:
            ::

                BytesFormat(b'Hello, %b!') % b'World'
                BytesFormat(b'%b, %b!') % (b'Hello', b'World')

        """
        fix_arg = self._fix_format_arg
        return self.how % tuple(fix_arg(item) for item in data)

    def join(self, *data: Iterable[MaybeBytes]) -> bytes:
        """Iterable join on a delimiter.

        Args:
            data: Iterable of items to join.

        Examples:
            ::

                BytesFormat(b' ').join([b'one', b'two', b'three'])

        """
        return self.how.join([bytes(item) for item in chain(*data)])


class ReView:
    """Wrap a :class:`~re.Pattern` so that it takes :class:`memoryview`
    strings.

    Note:
        This behavior is supported by :mod:`re`, but type stubs do not allow
        it. If that is ever resolved, remove this class.

    Args:
        pattern: The pattern to wrap.

    """

    def __init__(self, pattern: Pattern[bytes]) -> None:
        super().__init__()
        self._pattern = pattern

    @classmethod
    def compile(cls, pattern: bytes, *args, **kwargs) -> 'ReView':
        return cls(re.compile(pattern, *args, **kwargs))

    def match(self, string: Union[bytes, memoryview],
              *args, **kwargs) -> Optional[Match[bytes]]:
        string_b = cast(bytes, string)
        return self._pattern.match(string_b, *args, **kwargs)

    def fullmatch(self, string: Union[bytes, memoryview],
                  *args, **kwargs) -> Optional[Match[bytes]]:
        string_b = cast(bytes, string)
        return self._pattern.fullmatch(string_b, *args, **kwargs)

    def finditer(self, string: Union[bytes, memoryview],
                 *args, **kwargs) -> Iterable[Match[bytes]]:
        string_b = cast(bytes, string)
        return self._pattern.finditer(string_b, *args, **kwargs)

    def sub(self, repl: Union[bytes, Callable[[Match[bytes]], bytes]],
            string: Union[bytes, memoryview], *args, **kwargs) -> bytes:
        string_b = cast(bytes, string)
        return self._pattern.sub(repl, string_b, *args, **kwargs)


# Alias for ReView.
rev = ReView




from abc import abstractmethod
from argparse import Namespace, ArgumentParser
from asyncio import Task
from typing import Tuple
from typing_extensions import Protocol

from .session import LoginProtocol
from ..config import IMAPConfig

__all__ = ['BackendInterface', 'ServiceInterface']


class BackendInterface(Protocol):
    """Defines the abstract base class that is expected for backends that
    register themselves on the ``pymap.backend`` entry point.

    """

    @classmethod
    @abstractmethod
    def add_subparser(cls, subparsers) -> None:
        """Add a command-line argument sub-parser that will be used to choose
        this backend. For example::

            parser = subparsers.add_parser('foo', help='foo backend')
            parser.add_argument(...)

        Note:
            The name of the added sub-parser must be the same as the name of
            the entry point registered by the backend.

        Args:
            subparsers: The special action object as returned by
                :meth:`~argparse.ArgumentParser.add_subparsers`.

        """
        ...

    @classmethod
    @abstractmethod
    async def init(cls, args: Namespace) \
            -> Tuple['BackendInterface', IMAPConfig]:
        """Initialize the backend and return an instance.

        Args:
            args: The command-line arguments.

        """
        ...

    @property
    @abstractmethod
    def login(self) -> LoginProtocol:
        """Login callback that takes authentication credentials and returns a
        :class:`~pymap.interfaces.session.SessionInterface` object.

        """
        ...

    @property
    @abstractmethod
    def config(self) -> IMAPConfig:
        """The IMAP config in use by the backend."""
        ...


class ServiceInterface(Protocol):
    """Defines the abstract base class that is expected for services that
    register themselves on the ``pymap.service`` entry point.

    """

    @property
    @abstractmethod
    def task(self) -> Task:
        """The service's task that can waited or cancelled."""
        ...

    @classmethod
    @abstractmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        """Add the arguments or argument group used to configure the service.
        For example::

            group = parser.add_argument_group('foo service arguments')
            group.add_argument(...)

        Args:
            parser: The argument parser.

        """
        ...

    @classmethod
    @abstractmethod
    async def start(cls, backend: BackendInterface,
                    config: IMAPConfig) -> 'ServiceInterface':
        """Start the service and return the instance.

        Args:
            backend: The backend object.
            config: The config in use by the backend.

        """
        ...




import socket
from abc import abstractmethod, ABCMeta
from typing import Optional, Sequence, List

try:
    import systemd.daemon  # type: ignore
except ImportError as exc:
    systemd_import_exc: Optional[ImportError] = exc
else:
    systemd_import_exc = None

__all__ = ['InheritedSockets']


class InheritedSockets(metaclass=ABCMeta):
    """Abstracts the ability to retrieve inherited sockets from the calling
    process, usually a service management framework.

    """

    @abstractmethod
    def get(self) -> Sequence[socket.socket]:
        """Return the sockets inherited by the process."""
        ...

    @classmethod
    def of(cls, service_manager: str) -> 'InheritedSockets':
        """Return the inherited sockets for the given service manager.

        Note:
            Only ``'systemd'`` is implemented at this time.

        Args:
            service_manager: The service manager name.

        """
        if service_manager == 'systemd':
            return cls.for_systemd()
        else:
            raise KeyError(service_manager)

    @classmethod
    def for_systemd(cls) -> 'InheritedSockets':
        """Return the inherited sockets for `systemd`_. The `python-systemd`_
        library must be installed.

        See Also:
            `systemd.socket`_, `sd_listen_fds`_

        .. _systemd: https://freedesktop.org/wiki/Software/systemd/
        .. _systemd.socket: https://www.freedesktop.org/software/systemd/man/systemd.socket.html
        .. _sd_listen_fds: https://www.freedesktop.org/software/systemd/man/sd_listen_fds.html
        .. _python-systemd: https://github.com/systemd/python-systemd

        Raises:
            :exc:`NotImplementedError`

        """  # noqa: E501
        if systemd_import_exc:
            raise systemd_import_exc
        return _SystemdSockets()


class _SystemdSockets(InheritedSockets):

    def __init__(self) -> None:
        super().__init__()
        fds: Sequence[int] = systemd.daemon.listen_fds()
        sockets: List[socket.socket] = []
        for fd in fds:
            family = self._get_family(fd)
            sock = socket.fromfd(fd, family, socket.SOCK_STREAM)
            sockets.append(sock)
        self._sockets = sockets

    def get(self) -> Sequence[socket.socket]:
        return self._sockets

    @classmethod
    def _get_family(cls, fd: int) -> int:
        if systemd.daemon.is_socket(fd, socket.AF_UNIX):
            return socket.AF_UNIX
        elif systemd.daemon.is_socket(fd, socket.AF_INET6):
            return socket.AF_INET6
        else:
            return socket.AF_INET



"""Admin functions for a running pymap server."""

import asyncio
import sys
from argparse import ArgumentParser, Namespace, FileType
from typing import Type

from grpclib.client import Channel  # type: ignore
from pymap import __version__

from .append import AppendCommand
from .command import ClientCommand
from .. import AdminService
from ..grpc.admin_grpc import AdminStub


def main() -> int:
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('--version', action='version',
                        version='%(prog)s' + __version__)
    parser.add_argument('--outfile', metavar='PATH',
                        type=FileType('w'), default=sys.stdout,
                        help='the output file (default: stdout)')
    parser.add_argument('--socket', metavar='PATH', help='path to socket file')

    subparsers = parser.add_subparsers(dest='command',
                                       help='which admin command to run')
    commands = dict([AppendCommand.init(parser, subparsers)])
    args = parser.parse_args()

    if not args.command:
        parser.error('Expected command name.')
    command = commands[args.command]

    return asyncio.run(run(parser, args, command), debug=False)


async def run(parser: ArgumentParser, args: Namespace,
              command_cls: Type[ClientCommand]) -> int:
    loop = asyncio.get_event_loop()
    path = args.socket or AdminService.get_socket_path()
    channel = Channel(path=path, loop=loop)
    stub = AdminStub(channel)
    command = command_cls(stub, args)
    try:
        code = await command.run(args.outfile)
    finally:
        channel.close()
    return code
